#!/use/bin/tclsh
if {$argc < 3} {
    puts stderr "expected to get 3 arguments, VDD (volts), Length (nm), and name of BSIM model file"
    exit
}
set VDD [lindex $argv 0]
set LEN [lindex $argv 1]
set MOD [lindex $argv 2]
set DEBUG 0
######################################################################
# function performs transitions time extraction
#    inputs:
#	fname:	name of a ngspice-format output file
#	vlo:	voltage threshold for crossing
#	vhi:	voltage threshold for crossing
#    returns:
#	vals:	list of 1st rise or fall time for each waveform in file
#
proc transitions {fname vlo vhi {scale 1e12}} {
    set f [open $fname r]
    set d [read $f]
    close $f
# read the file
    set n 0
    foreach line [split $d "\n"] {
	if {[string length $line] == 0} continue
	if {[scan [lindex $line 0] %d N] > 0} {
	    set n [llength $line]
	    set time($N) [expr [lindex $line 1]*$scale]
	    for {set i 1} {$i < $n} {} {
		set data($N,$i) [lindex $line [incr i]]
	    }
	}
    }
    if {$n > 0} {
	set nvar [expr $n-2]
    } else {
	error "file $fname has no data"
    }
# calculate the crossings
    set vals {}
    for {set idx 1} {$idx <= $nvar} {incr idx} {
        set val -1.0
	set temp {}
	set tlo -1
	set thi -1
        for {set j 0; set i 1} {$i <= $N} {incr i; incr j} {
            set v1 [expr $data($i,$idx)-$vlo]
            set v2 [expr $data($j,$idx)-$vlo]
            if {[expr $v1*$v2] <= 0.0} {
                set tlo [expr $time($j)-$v2*($time($i)-$time($j))/($v1-$v2)]
	    }
            set v1 [expr $data($i,$idx)-$vhi]
            set v2 [expr $data($j,$idx)-$vhi]
            if {[expr $v1*$v2] <= 0.0} {
                set thi [expr $time($j)-$v2*($time($i)-$time($j))/($v1-$v2)]
	    }
	}
	if {$tlo == -1 || $thi == -1} {
	    error "waveform $idx did not switch!"
	} else {
	    set delta [expr abs($tlo-$thi)]
	}
	lappend vals $delta
    }
    return $vals
}
######################################################################
#   VDD	    power supply in voltes
#   LEN	    device length (nm)
#   PtoN    ratio of P to N device width
#   MODEL   name of BSIM model file
#
proc simulate {VDD LEN PtoN MODEL } {
    global DEBUG
# compute device dimensions
    set WN [expr $LEN*8]
    set WP [expr $LEN*8*$PtoN]
# create spice file, a sequence of 7 inverters fed by a single pulse, and
# loaded with a 20fF capacitor. print the waveforms at the innermost two
# inverters, one will be rising, and the other falling
    set f [open x.spice w]
    puts $f "* spice file
.include $MODEL
vdd  vdd  0  $VDD
vin  v0   0  0.0 pwl(0n 0,  10p 0,  60p $VDD,  300p $VDD)
.subckt  inv in out vdd
m1  out  in  vdd  vdd  pmos  l=${LEN}nm  w=${WP}nm
m2  out  in  0    0    nmos  l=${LEN}nm  w=${WN}nm
.ends inv
x1  v0  v1  vdd  inv
x2  v1  v2  vdd  inv
x3  v2  v3  vdd  inv
x4  v3  v4  vdd  inv
x5  v4  v5  vdd  inv
x6  v5  v6  vdd  inv
x7  v6  v7  vdd  inv
c1  v7  0  20fF
.opti nopage
.width out=240
.tran 0.2p 300p 0.0 0.2p
.print tran v(v4), v(v5)
.end"
    close $f
# run NGSPICE
    file delete x.wave
    catch {exec ngspice -b x.spice > x.wave} msg
    if {$DEBUG} {
	puts "ngspice result: $msg"
    }
# calculate the 20% to 80% transition times
    set result [transitions x.wave [expr 0.2*$VDD] [expr 0.8*$VDD]]
    set R0 [lindex $result 0]
    set R1 [lindex $result 1]
# we return the difference between the two transition times, this is the
# quantity we want to minimize!
    return [expr $R0 - $R1]
}
######################################################################
# utility procedure checks the signs of its two parameters
proc same_sign { x y } {
    if {$x >= 0.0 && $y >= 0.0} {return 1}
    if {$x <= 0.0 && $y <= 0.0} {return 1}
    return 0
}
######################################################################
# do a binary bisection search to find the point with the best match
# start with a wide range for the ratio, from 0.5 to 8!
set XLO 0.5
set YLO [simulate $VDD $LEN $XLO $MOD]
set XHI 8.0
set YHI [simulate $VDD $LEN $XHI $MOD]
# make sure we are OK in the beginning
if {[same_sign $YLO $YHI]} {
    error "$XLO and $XHI do not bracket the zero point!"
}
# do the iteration... until the ratio is to within 0.02 (about 1%)
while {($XHI-$XLO) > 0.02} {
    set XMID [expr 0.5*($XLO+$XHI)]
    set YMID [simulate $VDD $LEN $XMID $MOD]
    if {$DEBUG} {
	puts [format "%6.4f %6.3f -- %6.4f %6.3f -- %6.4f %6.3f" \
		$XLO $YLO $XMID $YMID $XHI $YHI]
    }
    if {[same_sign $YLO $YMID]} {
	set XLO $XMID
	set YLO $YMID
    } else {
	set XHI $XMID
	set YHI $YMID
    }
}
# we are done
puts "P to N ratio = [expr ($XHI+$XLO)*0.5]"

exit
